@@ -465,11 +465,16 @@ class CollectionClient(Client):
         temp_frame_prefix = "al_{}_collection_campaign_candidate_frames_{}_".format(
             current_time.strftime("%Y%m%d_%H%M%S_%f"), batch_uuid
         )
+
+        frame_dicts = []
+        for frame in frames:
+            frame_dict = frame.to_dict()
+            frame_dict["label_data"] = frame._labels.to_dict()["label_data"]
+            frame_dicts.append(frame_dict)
+
         frame_path = self._save_content_to_temp(
             temp_frame_prefix,
-            lambda x: self._write_frames_to_file(
-                [frame.to_dict() for frame in frames], x
-            ),
+            lambda x: self._write_frames_to_file(frame_dicts, x),
         )
         self.frame_batch_uuid_to_temp_file_path[batch_uuid] = frame_path
 
