@@ -15,6 +15,7 @@ from typing import Any, Optional, Union, Callable, List, Dict, Tuple, Set, IO, c
 from typing_extensions import Protocol, TypedDict
 
 from .util import (
+    ResizeMode,
     _is_one_gb_available,
     assert_valid_name,
     add_object_user_attrs,
@@ -32,6 +33,7 @@ from .util import (
     CropEmbeddingDict,
     LabelType,
     LabelFrameSummary,
+    DUMMY_FRAME_EMBEDDING,
 )
 from .frames import BaseFrame
 from .labels import GTLabelSet, InferenceLabelSet
@@ -59,12 +61,18 @@ class LabeledFrame(BaseFrame):
         frame_id: str,
         date_captured: Optional[str] = None,
         device_id: Optional[str] = None,
+        reuse_latest_embedding: Optional[bool] = False,
     ) -> None:
         super(LabeledFrame, self).__init__(
-            frame_id=frame_id, date_captured=date_captured, device_id=device_id
+            frame_id=frame_id,
+            date_captured=date_captured,
+            device_id=device_id,
+            reuse_latest_embedding=reuse_latest_embedding,
         )
 
-        self._labels = GTLabelSet(frame_id=frame_id)
+        self._labels = GTLabelSet(
+            frame_id=frame_id, reuse_latest_embedding=reuse_latest_embedding
+        )
 
     def _all_label_classes(self) -> List[str]:
         return self._labels._all_label_classes()
@@ -146,6 +154,7 @@ class LabeledFrame(BaseFrame):
         token: str,
         classification: str,
         visible: bool,
+        confidence: Optional[float] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
     ) -> None:
         """Add a label for a text token.
@@ -157,6 +166,7 @@ class LabeledFrame(BaseFrame):
             token (str): the text content of this token
             classification (str): the classification string
             visible (bool): is this a visible token in the text
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
         """
         self._labels.add_text_token(
@@ -166,7 +176,7 @@ class LabeledFrame(BaseFrame):
             token=token,
             classification=classification,
             visible=visible,
-            confidence=None,
+            confidence=confidence,
             user_attrs=user_attrs,
         )
 
@@ -179,6 +189,7 @@ class LabeledFrame(BaseFrame):
         position: List[float],
         dimensions: List[float],
         rotation: List[float],
+        confidence: Optional[float] = None,
         iscrowd: Optional[bool] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
         links: Optional[Dict[str, Any]] = None,
@@ -192,18 +203,26 @@ class LabeledFrame(BaseFrame):
             position (list of float): the position of the center of the cuboid
             dimensions (list of float): the dimensions of the cuboid
             rotation (list of float): the local rotation of the cuboid, represented as an xyzw quaternion.
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
             links (dict, optional): Links between labels. Defaults to None.
             coord_frame_id (str, optional): Coordinate frame id. Defaults to 'world'
         """
+        if (
+            coord_frame_id
+            and coord_frame_id != "world"
+            and not self._coord_frame_exists(coord_frame_id)
+        ):
+            raise Exception(f"Coordinate frame {coord_frame_id} does not exist.")
+
         self._labels.add_3d_cuboid(
             label_id=label_id,
             classification=classification,
             position=position,
             dimensions=dimensions,
             rotation=rotation,
-            confidence=None,
+            confidence=confidence,
             iscrowd=iscrowd,
             user_attrs=user_attrs,
             links=links,
@@ -220,6 +239,7 @@ class LabeledFrame(BaseFrame):
         left: Union[int, float],
         width: Union[int, float],
         height: Union[int, float],
+        confidence: Optional[float] = None,
         iscrowd: Optional[bool] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
         links: Optional[Dict[str, Any]] = None,
@@ -234,6 +254,7 @@ class LabeledFrame(BaseFrame):
             left (int or float): The left of the box in pixels
             width (int or float): The width of the box in pixels
             height (int or float): The height of the box in pixels
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
             links (dict, optional): Links between labels. Defaults to None.
@@ -246,7 +267,7 @@ class LabeledFrame(BaseFrame):
             left=left,
             width=width,
             height=height,
-            confidence=None,
+            confidence=confidence,
             area=None,
             iscrowd=iscrowd,
             user_attrs=user_attrs,
@@ -263,6 +284,7 @@ class LabeledFrame(BaseFrame):
         y1: Union[int, float],
         x2: Union[int, float],
         y2: Union[int, float],
+        confidence: Optional[float] = None,
         iscrowd: Optional[bool] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
         links: Optional[Dict[str, Any]] = None,
@@ -277,6 +299,7 @@ class LabeledFrame(BaseFrame):
             y1 (int or float): The x-coord of the first vertex in pixels
             x2 (int or float): The x-coord of the first vertex in pixels
             y2 (int or float): The x-coord of the first vertex in pixels
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
         """
@@ -288,7 +311,7 @@ class LabeledFrame(BaseFrame):
             y1=y1,
             x2=x2,
             y2=y2,
-            confidence=None,
+            confidence=confidence,
             iscrowd=iscrowd,
             user_attrs=user_attrs,
             links=links,
@@ -305,6 +328,7 @@ class LabeledFrame(BaseFrame):
         width: Union[int, float],
         height: Union[int, float],
         keypoints: List[Dict[KEYPOINT_KEYS, Union[int, float, str]]],
+        confidence: Optional[float] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
     ) -> None:
         """Add a label for a 2D keypoints task.
@@ -323,6 +347,7 @@ class LabeledFrame(BaseFrame):
             width (int or float): The width of the box in pixels
             height (int or float): The height of the box in pixels
             keypoints (list of dicts): The keypoints of this detection
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
         """
         self._labels.add_2d_keypoints(
@@ -334,7 +359,7 @@ class LabeledFrame(BaseFrame):
             width=width,
             height=height,
             keypoints=keypoints,
-            confidence=None,
+            confidence=confidence,
             user_attrs=user_attrs,
         )
 
@@ -345,6 +370,7 @@ class LabeledFrame(BaseFrame):
         label_id: str,
         classification: str,
         polygons: List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]],
+        confidence: Optional[float] = None,
         center: Optional[List[Union[int, float]]] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
     ) -> None:
@@ -371,6 +397,7 @@ class LabeledFrame(BaseFrame):
             label_id (str): label_id which is unique across datasets and inferences.
             classification (str): the classification string
             polygons (list of dicts): The polygon geometry
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             center (list of ints or floats, optional): The center point of the instance
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
         """
@@ -379,11 +406,52 @@ class LabeledFrame(BaseFrame):
             label_id=label_id,
             classification=classification,
             polygons=polygons,
-            confidence=None,
+            confidence=confidence,
             center=center,
             user_attrs=user_attrs,
         )
 
+    def add_label_2d_instance_seg(
+        self,
+        *,
+        sensor_id: str,
+        label_id: str,
+        mask_url: Optional[str] = None,
+        mask_data: Optional[Any] = None,
+        instance_mapping: List[Dict[str, Any]],
+        resize_mode: Optional[ResizeMode] = None,
+    ) -> None:
+        """Add a label for 2D instance segmentation.
+
+        Args:
+            sensor_id (str): sensor_id
+            label_id (str): label_id which is unique across datasets and inferences.
+            mask_url (str, optional): URL to the pixel mask png.
+            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
+            instance_mapping (List[Dict]): a list of instances present in the mask,
+                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
+                As an example of one instance:
+                .. code-block::
+                        {
+                            'id': 1,
+                            'classification': "Person",
+                            'attributes': {
+                                'is_standing': false,
+                            }
+                        }
+            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
+                "fill": stretch the mask to fit the base image dimensions
+                None: no change
+        """
+        self._labels.add_2d_instance_seg(
+            sensor_id=sensor_id,
+            label_id=label_id,
+            mask_url=mask_url,
+            mask_data=mask_data,
+            instance_mapping=instance_mapping,
+            resize_mode=resize_mode,
+        )
+
     def add_label_2d_semseg(
         self,
         *,
@@ -391,6 +459,7 @@ class LabeledFrame(BaseFrame):
         label_id: str,
         mask_url: Optional[str] = None,
         mask_data: Optional[Any] = None,
+        resize_mode: Optional[ResizeMode] = None,
     ) -> None:
         """Add a label for 2D semseg.
 
@@ -399,12 +468,16 @@ class LabeledFrame(BaseFrame):
             label_id (str): label_id which is unique across datasets and inferences.
             mask_url (str, optional): URL to the pixel mask png.
             mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
+            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
+                "fill": stretch the mask to fit the base image dimensions
+                None: no change
         """
         self._labels.add_2d_semseg(
             sensor_id=sensor_id,
             label_id=label_id,
             mask_url=mask_url,
             mask_data=mask_data,
+            resize_mode=resize_mode,
         )
 
     # TODO: handle secondary labels
@@ -414,6 +487,7 @@ class LabeledFrame(BaseFrame):
         sensor_id: str,
         label_id: str,
         classification: str,
+        confidence: Optional[float] = None,
         secondary_labels: Optional[Dict[str, Any]] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
     ) -> None:
@@ -423,6 +497,7 @@ class LabeledFrame(BaseFrame):
             sensor_id (str): sensor_id
             label_id (str): label_id which is unique across datasets and inferences.
             classification (str): the classification string
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             secondary_labels (dict, optional): dictionary of secondary labels
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
         """
@@ -431,7 +506,7 @@ class LabeledFrame(BaseFrame):
             label_id=label_id,
             classification=classification,
             secondary_labels=secondary_labels,
-            confidence=None,
+            confidence=confidence,
             user_attrs=user_attrs,
         )
 
@@ -440,6 +515,7 @@ class LabeledFrame(BaseFrame):
         *,
         label_id: str,
         classification: str,
+        confidence: Optional[float] = None,
         coord_frame_id: Optional[str] = None,
         user_attrs: Optional[Dict[str, Any]] = None,
     ) -> None:
@@ -448,13 +524,21 @@ class LabeledFrame(BaseFrame):
         Args:
             label_id (str): label_id which is unique across datasets and inferences.
             classification (str): the classification string
+            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
             coord_frame_id (str, optional): The coordinate frame id.
             user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
         """
+        if (
+            coord_frame_id
+            and coord_frame_id != "world"
+            and not self._coord_frame_exists(coord_frame_id)
+        ):
+            raise Exception(f"Coordinate frame {coord_frame_id} does not exist.")
+
         self._labels.add_3d_classification(
             label_id=label_id,
             classification=classification,
-            confidence=None,
+            confidence=confidence,
             coord_frame_id=coord_frame_id,
             user_attrs=user_attrs,
         )
@@ -490,6 +574,8 @@ class LabeledDataset:
     _temp_crop_embeddings_file_names_streaming: List[str]
     pipeline_mode: str
     backfilled_crops_flag: bool
+    backfilled_frame_embs_flag: bool
+    labels_with_confidence: bool
 
     def __init__(self, *, pipeline_mode: str = "BATCH") -> None:
         self._frames = []
@@ -525,6 +611,8 @@ class LabeledDataset:
         self._temp_crop_embeddings_file_names_streaming = []
         self.pipeline_mode = pipeline_mode
         self.backfilled_crops_flag = False
+        self.backfilled_frame_embs_flag = False
+        self.labels_with_confidence = False
 
     def _cleanup_temp_dir(self) -> None:
         mark_temp_directory_complete(self.temp_file_path)
@@ -548,6 +636,18 @@ class LabeledDataset:
                 loaded: Dict[str, Any] = json.loads(first_frame_json)
                 return loaded
 
+    def _check_labels_for_confidence(self, frame: LabeledFrame) -> None:
+        """Checks if any GT labels have confidence and sets a flag if this is the case.
+        (Only unlabeled dataset labels can have confidence values).
+        """
+        if self.labels_with_confidence:
+            return
+
+        for label in frame._labels.label_data:
+            if label.get("attributes", {}).get("confidence") is not None:
+                self.labels_with_confidence = True
+                return
+
     def _save_rows_to_temp(
         self,
         file_name_prefix: str,
@@ -583,8 +683,30 @@ class LabeledDataset:
         data_rows_content.close()
         return data_rows_content_path
 
+    def _backfill_frame_embeddings_if_necessary(self) -> bool:
+        """Determines if there are provided crop embeddings but no frame embeddings
+        Backfills frame embeddings with dummy [0.0, 0.0, 0.0] if crop embeddings exist
+        Returns True if backfill was necessary
+
+        Returns:
+            bool: returns if backfill on frame was necessary
+        """
+        if all([len(frame._labels.crop_embeddings) == 0 for frame in self._frames]):
+            return False
+
+        count = len([frame for frame in self._frames if frame.embedding is not None])
+        if count != 0:
+            return False
+
+        for frame in self._frames:
+            frame.add_frame_embedding(embedding=DUMMY_FRAME_EMBEDDING)
+
+        return True
+
     def _flush_to_disk(self) -> None:
         """Writes the all the frames in the frame buffer to temp file on disk"""
+        self.backfilled_frame_embs_flag = self._backfill_frame_embeddings_if_necessary()
+
         if self.pipeline_mode == "STREAMING":
             self._flush_to_disk_streaming()
         else:
@@ -690,6 +812,25 @@ class LabeledDataset:
                     f"Frame {frame_id} {issue_text}. Dataset frames for 2D_SEMSEG projects must have exactly one 2d_semseg label"
                 )
 
+        if primary_task == "2D_INSTANCE_SEGMENTATION":
+            label_counts = frame_summary["label_counts"]
+            url_count = label_counts.get("INSTANCE_LABEL_URL_2D", 0)
+            assets_count = label_counts.get("INSTANCE_LABEL_ASSET_2D", 0)
+            if url_count + assets_count != 1:
+                extra_labels = filter(
+                    lambda x: x != "INSTANCE_LABEL_URL_2D"
+                    and x != "INSTANCE_LABEL_ASSET_2D",
+                    label_counts.keys(),
+                )
+                issue_text = (
+                    "has no labels"
+                    if not len(label_counts)
+                    else f"has label types of {list(extra_labels)}"
+                )
+                raise Exception(
+                    f"Frame {frame_id} {issue_text}. Dataset frames for 2D_INSTANCE_SEGMENTATION projects must have exactly one 2d_instance_seg label"
+                )
+
     def _validate_frames(self, project_info: Dict[str, Any]) -> None:
         """Validates all frames in set according to project constraints
 
@@ -702,6 +843,10 @@ class LabeledDataset:
             print(
                 "WARNING: One or more Frames contained Frame Embeddings with Crops with no Crop Embeddings. All crops without Crop Embeddings have been backfilled with a default Embedding Vector"
             )
+        if self.backfilled_frame_embs_flag:
+            print(
+                "WARNING: Crop Embeddings were provided but no Frame Embeddings were provided. All frames have backfilled with a default Frame Embedding Vector"
+            )
 
     def add_frame(self, frame: LabeledFrame) -> None:
         """Add a LabeledFrame to this dataset.
@@ -722,6 +867,7 @@ class LabeledDataset:
             )
 
         self.backfilled_crops_flag = frame.validate_and_backfill_crop_embeddings()
+        self._check_labels_for_confidence(frame)
         self._frames.append(frame)
         self._frame_ids_set.add(frame.frame_id)
         self._label_ids_set.update(frame._labels._label_ids_set)
