@@ -1,5 +1,5 @@
 import requests
-from typing import Any, Union, Iterable, Dict, List
+from typing import Any, Union, Iterable, Dict, List, Callable
 from typing_extensions import Literal
 from google.cloud import storage  # type: ignore
 from http import HTTPStatus
@@ -9,14 +9,22 @@ from http import HTTPStatus
 UrlCheckResult = Literal["SKIPPED", "FAILURE", "SUCCESS", "S3_PATH", "INVALID_URL"]
 
 
-def check_urls(urls: Iterable[str], gs_client: Any) -> Dict[str, UrlCheckResult]:
+def check_urls(
+    urls: Iterable[str],
+    gs_client: Any,
+    requesting_func: Callable[..., requests.Response] = requests.get,
+) -> Dict[str, UrlCheckResult]:
     result = {}
     for url in urls:
-        result[url] = check_url(url, gs_client)
+        result[url] = check_url(url, gs_client, requesting_func=requesting_func)
     return result
 
 
-def check_url(url: str, gs_client: Any) -> UrlCheckResult:
+def check_url(
+    url: str,
+    gs_client: Any,
+    requesting_func: Callable[..., requests.Response] = requests.get,
+) -> UrlCheckResult:
     if url.lower().startswith("gs://"):
         if gs_client is None:
             return "SKIPPED"
@@ -34,11 +42,13 @@ def check_url(url: str, gs_client: Any) -> UrlCheckResult:
                 return "FAILURE"
     elif url.lower().startswith("http"):
         try:
-            code = requests.get(url, timeout=5).status_code
-            if code == HTTPStatus.OK:
-                return "SUCCESS"
-            else:
-                return "FAILURE"
+            # we use stream=True to avoid downloading the entire file for SSRF check
+            with requesting_func(url, timeout=5, stream=True) as response:
+                code = response.status_code
+                if code == HTTPStatus.OK:
+                    return "SUCCESS"
+                else:
+                    return "FAILURE"
         except:
             return "FAILURE"
     # We support S3 paths but only the user can access them from the app, and only if they provide their S3 creds
