@@ -16,10 +16,18 @@ from typing import Any, Optional, Union, Callable, List, Dict, Tuple, Set, IO, c
 from typing_extensions import Protocol, TypedDict
 
 from .util import (
+    ALL_CAMERA_MODELS,
+    ALL_ORIENTATION_KEYS,
+    ALL_POSITION_KEYS,
+    CAMERA_MODELS,
+    ResizeMode,
     _is_one_gb_available,
+    assert_and_get_valid_iso_date,
     assert_valid_name,
     add_object_user_attrs,
     create_temp_directory,
+    is_valid_float,
+    is_valid_number,
     mark_temp_directory_complete,
     TYPE_PRIMITIVE_TO_STRING_MAP,
     USER_METADATA_SEQUENCE,
@@ -75,6 +83,7 @@ class BaseFrame:
     _coord_frame_ids_set: Set[str]
     embedding: Optional[FrameEmbeddingDict]
     _label_ids_set: Set[str]
+    reuse_latest_embedding: Optional[bool]
 
     def __init__(
         self,
@@ -82,6 +91,7 @@ class BaseFrame:
         frame_id: str,
         date_captured: Optional[str] = None,
         device_id: Optional[str] = None,
+        reuse_latest_embedding: Optional[bool] = False,
     ) -> None:
         if not isinstance(frame_id, str):
             raise Exception("frame ids must be strings")
@@ -92,7 +102,7 @@ class BaseFrame:
         self.frame_id = frame_id
 
         if date_captured is not None:
-            self.date_captured = date_captured
+            self.date_captured = assert_and_get_valid_iso_date(date_captured)
         else:
             self.date_captured = str(datetime.datetime.now())
 
@@ -106,6 +116,7 @@ class BaseFrame:
         self.geo_data = {}
         self.user_metadata = []
         self.embedding = None
+        self.reuse_latest_embedding = reuse_latest_embedding
 
         self._coord_frame_ids_set = set()
         self._label_ids_set = set()
@@ -196,6 +207,13 @@ class BaseFrame:
         if val_type is None:
             val_type = TYPE_PRIMITIVE_TO_STRING_MAP[type(val)]
 
+        # If type is float, ensure that it is a valid float
+        if val is not None and val_type == "float" and not is_valid_float(val):
+            raise Exception(
+                f"For frame_id {self.frame_id}, metadata key: {key}, value: {val}, "
+                f"type is inferred as float but is not a valid float"
+            )
+
         # Checks that inferred type matches what the user put in val_type
         if val is not None:
             for (
@@ -290,6 +308,15 @@ class BaseFrame:
             inferred_val_type = type(val[0])
             list_elt_type = TYPE_PRIMITIVE_TO_STRING_MAP[inferred_val_type]
 
+        # If element type is float, ensure that all elements are valid floats
+        if val is not None and list_elt_type == "float":
+            for val_elt in val:
+                if not is_valid_float(val_elt):
+                    raise Exception(
+                        f"For frame_id {self.frame_id}, metadata key: {key}, value: {val}, "
+                        f"element type is inferred as float but is not a valid float"
+                    )
+
         self.user_metadata.append((key, val, list_elt_type, "list"))
 
     def add_geo_latlong_data(self, lat: float, lon: float) -> None:
@@ -316,6 +343,7 @@ class BaseFrame:
         pcd_url: str,
         coord_frame_id: Optional[str] = None,
         date_captured: Optional[str] = None,
+        mirror_asset: Optional[bool] = False,
     ) -> None:
         """Add a point cloud sensor data point to this frame,
         contained in PCD format. ascii, binary, and binary_compressed formats are supported.
@@ -327,6 +355,7 @@ class BaseFrame:
             pcd_url (str): URL to PCD formated data
             coord_frame_id (Optional[str], optional): The coordinate frame id. Defaults to None.
             date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
+            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed url assets.
         """
         if not isinstance(sensor_id, str):
             raise Exception("sensor ids must be strings")
@@ -334,6 +363,8 @@ class BaseFrame:
         if date_captured is None:
             date_captured = str(datetime.datetime.now())
 
+        date_captured = assert_and_get_valid_iso_date(date_captured)
+
         if coord_frame_id is None:
             coord_frame_id = "world"
 
@@ -362,6 +393,7 @@ class BaseFrame:
                 "sensor_id": sensor_id,
                 "sensor_metadata": {},
                 "sensor_type": "POINTCLOUD_PCD_V0",
+                "mirror_asset": mirror_asset,
             }
         )
 
@@ -375,6 +407,7 @@ class BaseFrame:
         range_url: str = None,
         coord_frame_id: Optional[str] = None,
         date_captured: Optional[str] = None,
+        mirror_asset: Optional[bool] = False,
     ) -> None:
         """Add a point cloud sensor data point to this frame, contained in dense binary files of
         little-endian values, similar to the raw format of KITTI lidar data. You can provide
@@ -389,6 +422,7 @@ class BaseFrame:
             range_url (Optional[str]): URL for the Range Pointcloud: float32 [r1, r2, ...]
             coord_frame_id (Optional[str], optional): Id for the Coordinate Frame. Defaults to None.
             date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
+            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
         """
         if not isinstance(sensor_id, str):
             raise Exception("sensor ids must be strings")
@@ -406,6 +440,8 @@ class BaseFrame:
         if date_captured is None:
             date_captured = str(datetime.datetime.now())
 
+        date_captured = assert_and_get_valid_iso_date(date_captured)
+
         if coord_frame_id is None:
             coord_frame_id = "world"
 
@@ -440,6 +476,7 @@ class BaseFrame:
                 "sensor_id": sensor_id,
                 "sensor_metadata": {},
                 "sensor_type": "POINTCLOUD_V0",
+                "mirror_asset": mirror_asset,
             }
         )
 
@@ -450,6 +487,7 @@ class BaseFrame:
         obj_url: str,
         coord_frame_id: Optional[str] = None,
         date_captured: Optional[str] = None,
+        mirror_asset: Optional[bool] = False,
     ) -> None:
         """Add a .obj file to the frame for text based geometry
 
@@ -458,6 +496,7 @@ class BaseFrame:
             obj_url (str): URL to where the object is located
             coord_frame_id (Optional[str], optional): ID for the coordinate frame. Defaults to None.
             date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
+            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
         """
         if not isinstance(sensor_id, str):
             raise Exception("sensor ids must be strings")
@@ -465,6 +504,8 @@ class BaseFrame:
         if date_captured is None:
             date_captured = str(datetime.datetime.now())
 
+        date_captured = assert_and_get_valid_iso_date(date_captured)
+
         if coord_frame_id is None:
             coord_frame_id = "world"
 
@@ -488,6 +529,7 @@ class BaseFrame:
                 "sensor_id": sensor_id,
                 "sensor_metadata": {},
                 "sensor_type": "OBJ_V0",
+                "mirror_asset": mirror_asset,
             }
         )
 
@@ -500,6 +542,8 @@ class BaseFrame:
         date_captured: Optional[str] = None,
         width: Optional[int] = None,
         height: Optional[int] = None,
+        coord_frame_id: Optional[str] = None,
+        mirror_asset: Optional[bool] = False,
     ) -> None:
         """Add an image "sensor" data to this frame.
 
@@ -510,6 +554,8 @@ class BaseFrame:
             date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
             width (Optional[int], optional): The width of the image in pixels. Defaults to None.
             height (Optional[int], optional): The height of the image in pixels. Defaults to None.
+            coord_frame_id (Optional[str], optional): Id for the Coordinate Frame (only accepts coordinate frames of type "IMAGE" or "IMAGE_PROJECTION"). Defaults to None.
+            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
         """
         if not isinstance(sensor_id, str):
             raise Exception("sensor ids must be strings")
@@ -528,21 +574,47 @@ class BaseFrame:
         if date_captured is None:
             date_captured = str(datetime.datetime.now())
 
+        date_captured = assert_and_get_valid_iso_date(date_captured)
+
         data_urls = {"image_url": image_url}
         if preview_url is not None:
             data_urls["preview_url"] = preview_url
 
-        self._add_coordinate_frame(
-            {"coordinate_frame_id": sensor_id, "coordinate_frame_type": "IMAGE"}
-        )
+        if coord_frame_id is None:
+            coord_frame_id = sensor_id
+            self._add_coordinate_frame(
+                {
+                    "coordinate_frame_id": coord_frame_id,
+                    "coordinate_frame_type": "IMAGE",
+                }
+            )
+        elif not self._coord_frame_exists(coord_frame_id):
+            raise Exception(
+                "Coordinate frame {} does not exist.".format(coord_frame_id)
+            )
+
+        for coordinate_frame in self.coordinate_frames:
+            coord_type = coordinate_frame["coordinate_frame_type"]
+            if (
+                coordinate_frame["coordinate_frame_id"] == coord_frame_id
+                and coord_type != "IMAGE"
+                and coord_type != "IMAGE_PROJECTION"
+            ):
+                raise Exception(
+                    'Coordinate frame {} is of type {} and must be either "IMAGE" or "IMAGE_PROJECTION".'.format(
+                        coord_frame_id, coord_type
+                    )
+                )
+
         self.sensor_data.append(
             {
-                "coordinate_frame": sensor_id,
+                "coordinate_frame": coord_frame_id,
                 "data_urls": data_urls,
                 "date_captured": date_captured,
                 "sensor_id": sensor_id,
                 "sensor_metadata": sensor_metadata,
                 "sensor_type": "IMAGE_V0",
+                "mirror_asset": mirror_asset,
             }
         )
 
@@ -555,6 +627,8 @@ class BaseFrame:
         date_captured: Optional[str] = None,
         width: Optional[int] = None,
         height: Optional[int] = None,
+        mirror_asset: Optional[bool] = False,
+        resize_mode: Optional[ResizeMode] = None,
     ) -> None:
         """Add an image overlay for the given "sensor" to this frame.
 
@@ -565,6 +639,10 @@ class BaseFrame:
             date_captured (Optional[str], optional): ISO formatted date. Defaults to None.
             width (Optional[int], optional): The width of the image in pixels. Defaults to None.
             height (Optional[int], optional): The height of the image in pixels. Defaults to None.
+            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
+            resize_mode (ResizeMode, optional): if the overlay is a different size from the base image, define how to display it
+                "fill": stretch the overlay to fit the base image dimensions
+                None: no change
         """
         if not isinstance(sensor_id, str):
             raise Exception("sensor ids must be strings")
@@ -572,7 +650,10 @@ class BaseFrame:
         if not isinstance(overlay_id, str):
             raise Exception("overlay ids must be strings")
 
-        sensor_metadata: Dict[str, Union[str, int]] = {"overlay_id": overlay_id}
+        sensor_metadata: Dict[str, Union[str, int]] = {
+            "overlay_id": overlay_id,
+            "resize_mode": str(resize_mode),
+        }
         if width is not None:
             if not isinstance(width, int):
                 raise Exception("width must be an int")
@@ -586,6 +667,8 @@ class BaseFrame:
         if date_captured is None:
             date_captured = str(datetime.datetime.now())
 
+        date_captured = assert_and_get_valid_iso_date(date_captured)
+
         data_urls = {"image_url": image_url}
 
         self._add_coordinate_frame(
@@ -599,6 +682,7 @@ class BaseFrame:
                 "sensor_id": sensor_id,
                 "sensor_metadata": sensor_metadata,
                 "sensor_type": "IMAGE_OVERLAY_V0",
+                "mirror_asset": mirror_asset,
             }
         )
 
@@ -608,6 +692,7 @@ class BaseFrame:
         sensor_id: str,
         audio_url: str,
         date_captured: Optional[str] = None,
+        mirror_asset: Optional[bool] = False,
     ) -> None:
         """Add an audio "sensor" data to this frame.
 
@@ -615,6 +700,7 @@ class BaseFrame:
             sensor_id (str): The id of this sensor.
             audio_url (str): The URL to load this audio data (mp3, etc.).
             date_captured (str, optional): ISO formatted date. Defaults to None.
+            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
         """
         if not isinstance(sensor_id, str):
             raise Exception("sensor ids must be strings")
@@ -623,6 +709,8 @@ class BaseFrame:
         if date_captured is None:
             date_captured = str(datetime.datetime.now())
 
+        date_captured = assert_and_get_valid_iso_date(date_captured)
+
         data_urls = {"audio_url": audio_url}
 
         self._add_coordinate_frame(
@@ -636,6 +724,7 @@ class BaseFrame:
                 "sensor_id": sensor_id,
                 "sensor_metadata": sensor_metadata,
                 "sensor_type": "AUDIO_V0",
+                "mirror_asset": mirror_asset,
             }
         )
 
@@ -645,6 +734,7 @@ class BaseFrame:
         sensor_id: str,
         video_url: str,
         date_captured: Optional[str] = None,
+        mirror_asset: Optional[bool] = False,
     ) -> None:
         """Add a video "sensor" data to this frame.
 
@@ -652,6 +742,7 @@ class BaseFrame:
             sensor_id (str): The id of this sensor.
             video_url (str): The URL to load this video data (mp4, webm, etc.).
             date_captured (str, optional): ISO formatted date. Defaults to None.
+            mirror_asset (Optional[str]), optional): request this asset in this frame to be saved to Aquarium Storage. Useful for Short duration signed URL assets.
         """
         if not isinstance(sensor_id, str):
             raise Exception("sensor ids must be strings")
@@ -660,6 +751,8 @@ class BaseFrame:
         if date_captured is None:
             date_captured = str(datetime.datetime.now())
 
+        date_captured = assert_and_get_valid_iso_date(date_captured)
+
         data_urls = {"video_url": video_url}
 
         self._add_coordinate_frame(
@@ -673,6 +766,7 @@ class BaseFrame:
                 "sensor_id": sensor_id,
                 "sensor_metadata": sensor_metadata,
                 "sensor_type": "VIDEO_V0",
+                "mirror_asset": mirror_asset,
             }
         )
 
@@ -713,13 +807,26 @@ class BaseFrame:
 
         if position is None:
             position = {"x": 0, "y": 0, "z": 0}
+        if type(position) is not dict:
+            raise Exception("position improperly formatted")
+        for position_key in ALL_POSITION_KEYS:
+            if not is_valid_number(position[position_key]):
+                raise Exception("position coordinates must be valid numbers")
 
         if orientation is None:
             orientation = {"w": 1, "x": 0, "y": 0, "z": 0}
+        if type(orientation) is not dict:
+            raise Exception("orientation improperly formatted")
+        for orientation_key in ALL_ORIENTATION_KEYS:
+            if not is_valid_number(orientation[orientation_key]):
+                raise Exception("orientation coordinates must be valid numbers")
 
         if parent_frame_id is None:
             parent_frame_id = "world"
 
+        if not isinstance(parent_frame_id, str):
+            raise Exception("parent_frame_id must be a string or None.")
+
         metadata = {
             "position": position,
             "orientation": orientation,
@@ -734,6 +841,194 @@ class BaseFrame:
             }
         )
 
+    def add_coordinate_frame_2d(
+        self,
+        *,
+        coord_frame_id: str,
+        fx: Union[int, float],
+        fy: Union[int, float],
+        camera_model: Optional[CAMERA_MODELS] = None,
+        position: Optional[Dict[POSITION_KEYS, Union[int, float]]] = None,
+        orientation: Optional[Dict[ORIENTATION_KEYS, Union[int, float]]] = None,
+        camera_matrix: Optional[List[List[Union[int, float]]]] = None,
+        k1: Optional[Union[int, float]] = None,
+        k2: Optional[Union[int, float]] = None,
+        k3: Optional[Union[int, float]] = None,
+        k4: Optional[Union[int, float]] = None,
+        k5: Optional[Union[int, float]] = None,
+        k6: Optional[Union[int, float]] = None,
+        p1: Optional[Union[int, float]] = None,
+        p2: Optional[Union[int, float]] = None,
+        cx: Optional[Union[int, float]] = None,
+        cy: Optional[Union[int, float]] = None,
+        s1: Optional[Union[int, float]] = None,
+        s2: Optional[Union[int, float]] = None,
+        s3: Optional[Union[int, float]] = None,
+        s4: Optional[Union[int, float]] = None,
+        skew: Optional[Union[int, float]] = None,
+        parent_frame_id: Optional[str] = None,
+    ) -> None:
+        """Add a 2D Coordinate Frame.
+
+        Args:
+            coord_frame_id (str): String identifier for this coordinate frame.
+            fx (Union[int, float]): focal length x in pixels.
+            fy (Union[int, float]): focal length y in pixels.
+            camera_model (Optional[CAMERA_MODELS], optional): Either "fisheye" for the fisheye model, or "brown_conrady" for the pinhole model with Brown-Conrady distortion. Defaults to "brown_conrady".
+            position (Optional[Dict[POSITION, Union[int, float]]], optional): Dict of the form {x, y, z}. Defaults to {x: 0, y: 0, z: 0}.
+            orientation (Optional[Dict[ORIENTATION, Union[int, float]]], optional): Quaternion rotation dict of the form {w, x, y, z}. Defaults to {x: 0, y: 0, z: 0, w: 1}.
+            camera_matrix (Optional[List[List[Union[int, float]]]], optional): 4x4 row major order camera matrix mapping 3d world space to camera space (x right, y down, z forward). Keep in mind, if you pass in the camera matrix it will stack on top of the position/orientation you pass in as well. This is only needed if you cannot properly represent your camera using the position/orientation parameters. Defaults to None.
+            cx (Optional[Union[int, float]], optional): optical center pixel x coordinate. Defaults to x center of image.
+            cy (Optional[Union[int, float]], optional): optical center pixel y coordinate. Defaults to y center of image.
+            k1 (Optional[Union[int, float]], optional): k1 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
+            k2 (Optional[Union[int, float]], optional): k2 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
+            k3 (Optional[Union[int, float]], optional): k3 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
+            k4 (Optional[Union[int, float]], optional): k4 radial distortion coefficient (Brown-Conrady, fisheye). Defaults to 0.
+            k5 (Optional[Union[int, float]], optional): k5 radial distortion coefficient (Brown-Conrady). Defaults to 0.
+            k6 (Optional[Union[int, float]], optional): k6 radial distortion coefficient (Brown-Conrady). Defaults to 0.
+            p1 (Optional[Union[int, float]], optional): p1 tangential distortion coefficient (Brown-Conrady). Defaults to 0.
+            p2 (Optional[Union[int, float]], optional): p2 tangential distortion coefficient (Brown-Conrady). Defaults to 0.
+            s1 (Optional[Union[int, float]], optional): s1 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
+            s2 (Optional[Union[int, float]], optional): s2 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
+            s3 (Optional[Union[int, float]], optional): s3 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
+            s4 (Optional[Union[int, float]], optional): s4 thin prism distortion coefficient (Brown-Conrady). Defaults to 0.
+            skew (Optional[Union[int, float]], optional): camera skew coefficient (fisheye). Defaults to 0.
+            parent_frame_id (Optional[str], optional): String id of the parent coordinate frame. Defaults to None.
+        """
+
+        if not isinstance(coord_frame_id, str):
+            raise Exception("coord_frame_id must be a string")
+
+        if coord_frame_id == "world":
+            raise Exception("coord_frame_id cannot be world")
+
+        if self._coord_frame_exists(coord_frame_id):
+            raise Exception("Coordinate frame already exists.")
+
+        # If world doesn't exist, make the world coordinate frame
+        if not self._coord_frame_exists("world"):
+            self._add_coordinate_frame(
+                {
+                    "coordinate_frame_id": "world",
+                    "coordinate_frame_type": "WORLD",
+                }
+            )
+
+        if position is None:
+            position = {"x": 0, "y": 0, "z": 0}
+        if type(position) is not dict:
+            raise Exception("position improperly formatted")
+        for position_key in ALL_POSITION_KEYS:
+            if not is_valid_number(position[position_key]):
+                raise Exception("position coordinates must be valid numbers")
+
+        if orientation is None:
+            orientation = {"w": 1, "x": 0, "y": 0, "z": 0}
+        if type(orientation) is not dict:
+            raise Exception("orientation improperly formatted")
+        for orientation_key in ALL_ORIENTATION_KEYS:
+            if not is_valid_number(orientation[orientation_key]):
+                raise Exception("orientation coordinates must be valid numbers")
+
+        if camera_matrix is not None:
+            if type(camera_matrix) != list:
+                raise Exception("camera matrix must be a python list of lists")
+            if len(camera_matrix) != 4:
+                raise Exception("camera matrix is not a properly formatted 4x4 matrix")
+            for row in camera_matrix:
+                if type(camera_matrix) != list:
+                    raise Exception("camera matrix must be a python list of lists")
+                if len(row) != 4:
+                    raise Exception(
+                        "camera matrix is not a properly formatted 4x4 matrix"
+                    )
+                for el in row:
+                    if not is_valid_number(el):
+                        raise Exception(
+                            "element within camera matrix not a valid float/int"
+                        )
+
+        if parent_frame_id is None:
+            parent_frame_id = "world"
+
+        if not isinstance(parent_frame_id, str):
+            raise Exception("parent_frame_id must be a string or None.")
+
+        if not is_valid_number(fx) or not is_valid_number(fy):
+            raise Exception("focal lengths are required and must be valid numbers")
+
+        if fx == 0 or fy == 0:
+            raise Exception("focal lengths cannot be 0")
+
+        if camera_model is None:
+            camera_model = "brown_conrady"
+
+        if not isinstance(camera_model, str) or camera_model not in ALL_CAMERA_MODELS:
+            raise Exception(
+                "invalid camera model, valid values are {}".format(ALL_CAMERA_MODELS)
+            )
+
+        for coord in [k1, k2, k3, k4, k5, k6, p1, p2, cx, cy, s1, s2, s3, s4, skew]:
+            if coord is not None and not is_valid_number(coord):
+                raise Exception(
+                    f"{coord} is not a valid 2D coordinate frame attribute. Must be float/int/None"
+                )
+
+        if (cx is None and cy is not None) or (cx is not None and cy is None):
+            raise Exception("optical centers must both be None or both be numbers")
+        if (cx is not None and cx <= 0) or (cy is not None and cy <= 0):
+            raise Exception("optical centers must be greater than 0")
+
+        metadata: Dict[str, Any] = {
+            "fx": fx,
+            "fy": fy,
+            "camera_model": camera_model,
+            "position": position,
+            "orientation": orientation,
+            "parent_frame_id": parent_frame_id,
+        }
+
+        if camera_matrix is not None:
+            metadata["camera_matrix"] = camera_matrix
+        if k1 is not None:
+            metadata["k1"] = k1
+        if k2 is not None:
+            metadata["k2"] = k2
+        if k3 is not None:
+            metadata["k3"] = k3
+        if k4 is not None:
+            metadata["k4"] = k4
+        if k5 is not None:
+            metadata["k5"] = k5
+        if k6 is not None:
+            metadata["k6"] = k6
+        if p1 is not None:
+            metadata["p1"] = p1
+        if p2 is not None:
+            metadata["p2"] = p2
+        if cx is not None:
+            metadata["cx"] = cx
+        if cy is not None:
+            metadata["cy"] = cy
+        if s1 is not None:
+            metadata["s1"] = s1
+        if s2 is not None:
+            metadata["s2"] = s2
+        if s3 is not None:
+            metadata["s3"] = s3
+        if s4 is not None:
+            metadata["s4"] = s4
+        if skew is not None:
+            metadata["skew"] = skew
+
+        self._add_coordinate_frame(
+            {
+                "coordinate_frame_id": coord_frame_id,
+                "coordinate_frame_type": "IMAGE_PROJECTION",
+                "coordinate_frame_metadata": json.dumps(metadata),
+            }
+        )
+
     def add_text(
         self, *, sensor_id: str, text: str, date_captured: Optional[str] = None
     ) -> None:
@@ -779,6 +1074,7 @@ class BaseFrame:
             "coordinate_frames": self.coordinate_frames,
             "sensor_data": self.sensor_data,
             "geo_data": self.geo_data,
+            "reuse_latest_embedding": self.reuse_latest_embedding,
         }
         user_metadata_types = {}
         user_metadata_modes = {}
