@@ -16,6 +16,8 @@ from typing_extensions import Protocol
 
 
 from .util import (
+    DUMMY_FRAME_EMBEDDING,
+    ResizeMode,
     _is_one_gb_available,
     create_temp_directory,
     mark_temp_directory_complete,
@@ -40,8 +42,12 @@ class InferenceFrameWithEmbeddings(Protocol):
 class InferencesFrame(InferenceLabelSet):
     embedding: Optional[FrameEmbeddingDict]
 
-    def __init__(self, *, frame_id: str) -> None:
-        super().__init__(frame_id=frame_id)
+    def __init__(
+        self, *, frame_id: str, reuse_latest_embedding: Optional[bool] = False
+    ) -> None:
+        super().__init__(
+            frame_id=frame_id, reuse_latest_embedding=reuse_latest_embedding
+        )
         self.embedding = None
 
     def add_frame_embedding(
@@ -344,6 +350,7 @@ class InferencesFrame(InferenceLabelSet):
         label_id: str,
         mask_url: Optional[str] = None,
         mask_data: Optional[Any] = None,
+        resize_mode: Optional[ResizeMode] = None,
     ) -> None:
         """Add an inference for 2D semseg.
 
@@ -352,12 +359,57 @@ class InferencesFrame(InferenceLabelSet):
             label_id (str): label_id which is unique across datasets and inferences.
             mask_url (str, optional): URL to the pixel mask png.
             mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
+            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
+                "fill": stretch the mask to fit the base image dimensions
+                None: no change
         """
         self.add_2d_semseg(
             sensor_id=sensor_id,
             label_id=label_id,
             mask_url=mask_url,
             mask_data=mask_data,
+            resize_mode=resize_mode,
+        )
+
+    def add_inference_2d_instance_seg(
+        self,
+        *,
+        sensor_id: str,
+        label_id: str,
+        mask_url: Optional[str] = None,
+        mask_data: Optional[Any] = None,
+        instance_mapping: List[Dict[str, Any]],
+        resize_mode: Optional[ResizeMode] = None,
+    ) -> None:
+        """Add an inference for 2D instance segmentation.
+
+        Args:
+            sensor_id (str): sensor_id
+            label_id (str): label_id which is unique across datasets and inferences.
+            mask_url (str, optional): URL to the pixel mask png.
+            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
+            instance_mapping (List[Dict]): a list of instances present in the mask,
+                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
+                As an example of one instance:
+                .. code-block::
+                        {
+                            'id': 1,
+                            'classification': "Person",
+                            'attributes': {
+                                'is_standing': false,
+                            }
+                        }
+            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
+                "fill": stretch the mask to fit the base image dimensions
+                None: no change
+        """
+        self.add_2d_instance_seg(
+            sensor_id=sensor_id,
+            label_id=label_id,
+            mask_url=mask_url,
+            mask_data=mask_data,
+            instance_mapping=instance_mapping,
+            resize_mode=resize_mode,
         )
 
     def add_inference_2d_classification(
@@ -429,6 +481,7 @@ class Inferences:
     _temp_frame_file_names: List[str]
     _temp_frame_asset_file_names: List[str]
     _temp_frame_embeddings_file_names: List[str]
+    backfilled_frame_embs_flag: bool
 
     def __init__(self) -> None:
         if not _is_one_gb_available():
@@ -455,6 +508,7 @@ class Inferences:
         self._temp_frame_file_names = []
         self._temp_frame_embeddings_file_names = []
         self._temp_frame_asset_file_names = []
+        self.backfilled_frame_embs_flag = False
 
     def _cleanup_temp_dir(self) -> None:
         mark_temp_directory_complete(self.temp_file_path)
@@ -494,8 +548,30 @@ class Inferences:
         data_rows_content.close()
         return data_rows_content_path
 
+    def _backfill_frame_embeddings_if_necessary(self) -> bool:
+        """Determines if there are provided crop embeddings but no frame embeddings
+        Backfills frame embeddings with dummy [0.0, 0.0, 0.0] if crop embeddings exist
+        Returns True if backfill was necessary
+
+        Returns:
+            bool: returns if backfill on frame was necessary
+        """
+        if all([len(inf_frame.crop_embeddings) == 0 for inf_frame in self._frames]):
+            return False
+
+        count = len([frame for frame in self._frames if frame.embedding is not None])
+        if count != 0:
+            return False
+
+        for frame in self._frames:
+            frame.add_frame_embedding(embedding=DUMMY_FRAME_EMBEDDING)
+
+        return True
+
     def _flush_to_disk(self) -> None:
         """Writes the all the frames in the frame buffer to temp file on disk"""
+        self.backfilled_frame_embs_flag = self._backfill_frame_embeddings_if_necessary()
+
         if len(self._frames) == 0:
             return
         frame_path = self._save_rows_to_temp(
@@ -555,6 +631,25 @@ class Inferences:
                     f"Frame {frame_id} {issue_text}. Inferences frames for 2D_SEMSEG projects must have exactly one 2d_semseg inference "
                 )
 
+        if primary_task == "2D_INSTANCE_SEGMENTATION":
+            label_counts = frame_summary["label_counts"]
+            url_count = label_counts.get("INSTANCE_LABEL_URL_2D", 0)
+            assets_count = label_counts.get("INSTANCE_LABEL_ASSET_2D", 0)
+            if url_count + assets_count != 1:
+                extra_labels = filter(
+                    lambda x: x != "INSTANCE_LABEL_URL_2D"
+                    and x != "INSTANCE_LABEL_ASSET_2D",
+                    label_counts.keys(),
+                )
+                issue_text = (
+                    "has no labels"
+                    if not len(label_counts)
+                    else f"has label types of {list(extra_labels)}"
+                )
+                raise Exception(
+                    f"Frame {frame_id} {issue_text}. Inferences frames for 2D_INSTANCE_SEGMENTATION projects must have exactly one 2d_instance_seg inference"
+                )
+
         project_custom_metrics_names = [
             custom_metric["name"]
             for custom_metric in project_info.get("custom_metrics", []) or []
@@ -594,6 +689,20 @@ class Inferences:
         for frame_summary in self._frame_summaries:
             self._validate_frame(frame_summary, project_info)
 
+        primary_task = project_info.get("primary_task")
+        if primary_task == "BINARY_CLASSIFICATION":
+            negative_class = project_info.get("binary_classification_negative_class")
+            classifications = self._label_classes_set
+            if negative_class in classifications:
+                raise Exception(
+                    f"Inferences must be submitted as the positive class with confidence 0 to 1 for binary classification project {project_info['id']}. Inferences will be interpreted by Aquarium using the decision threshold."
+                )
+
+        if self.backfilled_frame_embs_flag:
+            print(
+                "WARNING: Crop Embeddings were provided but no Frame Embeddings were provided. All inf frames have backfilled with a default Frame Embedding Vector"
+            )
+
     def add_frame(self, frame: InferencesFrame) -> None:
         """Add an InferencesFrame to this dataset.
 
