@@ -63,6 +63,7 @@ class IssueElement:
         frame_data: Optional[Dict[str, Any]] = None,
         crop_data: Optional[Dict[str, Any]] = None,
         label_metadata: Optional[Dict[str, Any]] = None,
+        comments: Optional[Dict[str, Any]] = None,
     ):
         if element_type != "crop" and element_type != "frame":
             raise Exception('element_type must be either "crop" or "frame"')
@@ -76,6 +77,7 @@ class IssueElement:
         self.frame_data = frame_data
         self.crop_data = crop_data
         self.label_metadata = label_metadata
+        self.comments = comments
 
     def to_dict(self) -> Dict[str, Any]:
         return {
@@ -88,6 +90,7 @@ class IssueElement:
             "frame_data": self.frame_data,
             "crop_data": self.crop_data,
             "label_metadata": self.label_metadata,
+            "comments": self.comments,
         }
 
     # For element modification
@@ -130,6 +133,7 @@ class IssueElementApiResp(TypedDict):
     frameData: Optional[Dict[str, Any]]
     cropData: Optional[Dict[str, Any]]
     labelMetadata: Optional[Dict[str, Any]]
+    comments: Optional[Dict[str, Any]]
 
 
 class IssueApiResp(TypedDict):
@@ -145,6 +149,9 @@ class IssueApiResp(TypedDict):
     state: Optional[IssueState]
     issue_id: Optional[str]
     elements: List[IssueElementApiResp]
+    comments: Optional[Dict[str, Any]]
+    template_type: Optional[str]
+    custom_template_fields: Optional[Dict[str, Any]]
 
 
 class Issue:
@@ -162,6 +169,8 @@ class Issue:
         state (str): Current state of issue ("triage", "inProgress", "inReview", "resolved", "cancelled"). Defaults to "triage".
         issue_id (str): The issue id.
         inference_set (Optional[str], optional): The inference set for this issue. Defaults to None.
+        template_type (Optional[str]): The template type for this issue. (read-only)
+        custom_template_fields (Optional[Dict[str, Any]]): Custom fields associated with the issue template type. (read-only)
     """
 
     def __init__(
@@ -178,6 +187,9 @@ class Issue:
         state: Optional[str] = None,
         issue_id: Optional[str] = None,
         inference_set: Optional[str] = None,
+        comments: Optional[Dict[str, Any]] = None,
+        template_type: Optional[str] = None,
+        custom_template_fields: Optional[Dict[str, Any]] = None,
     ):
 
         self.name = name
@@ -191,6 +203,9 @@ class Issue:
         self.state = state
         self.issue_id = issue_id
         self.inference_set = inference_set
+        self.comments = comments
+        self.template_type = template_type
+        self.custom_template_fields = custom_template_fields
 
     def __repr__(self) -> str:
         return "Issue {} ({})".format(self.issue_id, self.name)
@@ -211,6 +226,9 @@ class Issue:
             "state": self.state,
             "issue_id": self.issue_id,
             "inference_set": self.inference_set,
+            "comments": self.comments,
+            "template_type": self.template_type,
+            "custom_template_fields": self.custom_template_fields,
         }
 
 
@@ -260,6 +278,7 @@ class IssueManager:
                     frame_data=raw_el.get("frameData"),
                     crop_data=raw_el.get("cropData"),
                     label_metadata=raw_el.get("labelMetadata"),
+                    comments=raw_el.get("comments"),
                 )
             )
 
@@ -273,9 +292,12 @@ class IssueManager:
             assignee=api_resp.get("assignee"),
             state=api_resp.get("state"),
             issue_id=api_resp["id"],
+            comments=api_resp.get("comments"),
             dataset=dataset,
             inference_set=inference_set,
             elements=elements,
+            template_type=api_resp.get("template_type"),
+            custom_template_fields=api_resp.get("custom_template_fields"),
         )
 
     def add_elements_to_issue(
@@ -444,16 +466,22 @@ class IssueManager:
 
         return resp_data["id"]
 
-    def get_issue(self, issue_id: str, exclude_frame_data: bool = False) -> Issue:
+    def get_issue(
+        self,
+        issue_id: str,
+        exclude_frame_data: bool = False,
+        include_comment_data: bool = False,
+    ) -> Issue:
         """Get a specific issue in the associated project.
         This will also include all associated frame metadata associated with each element.
 
         Args:
             issue_id (str): The issue id.
             exclude_frame_data (bool): Set to True to exclude full frame data from the issue element (e.g. to cut down on download size).
+            include_comment_data (bool): Set to True to include comments on the issue and issue elements.
 
         Returns:
-            Issue: The issue data (including frame_data, crop_data, and label_metadata).
+            Issue: The issue data (including frame_data, crop_data, label_metadata, and comments).
         """
         url = "/projects/{}/issues/{}/download_elements".format(
             self.project_id, issue_id
@@ -462,7 +490,11 @@ class IssueManager:
         r = requests.get(
             self.client.api_endpoint + url,
             headers=self.client._get_creds_headers(),
-            params={"get_as_gcs_file": True, "exclude_frame_data": exclude_frame_data},
+            params={
+                "get_as_gcs_file": True,
+                "exclude_frame_data": exclude_frame_data,
+                "include_comment_data": include_comment_data,
+            },
         )
 
         raise_resp_exception_error(r)
