@@ -13,6 +13,7 @@ import pandas as pd
 from typing import Any, Optional, Union, Callable, List, Dict, Tuple, Set, IO, cast
 from typing_extensions import Literal
 from uuid import uuid4
+from .util import is_valid_number, is_valid_float
 
 HAS_PILLOW = False
 try:
@@ -26,6 +27,7 @@ except ImportError:
 from .util import (
     _is_one_gb_available,
     add_object_user_attrs,
+    assert_valid_name,
     create_temp_directory,
     mark_temp_directory_complete,
     POLYGON_VERTICES_KEYS,
@@ -39,6 +41,7 @@ from .util import (
     LabelFrameSummary,
     BaseLabelEntryDict,
     BaseLabelAssetDict,
+    ResizeMode,
 )
 
 CustomMetricsEntry = Union[float, List[List[Union[float, int]]]]
@@ -60,8 +63,15 @@ class BaseLabelSet:
     _label_ids_set: Set[str]
     crop_embeddings: List[CropEmbeddingDict]
     _seen_label_class_ids: Set[int]
+    reuse_latest_embedding: bool
 
-    def __init__(self, *, frame_id: str, crop_type: CropType) -> None:
+    def __init__(
+        self,
+        *,
+        frame_id: str,
+        crop_type: CropType,
+        reuse_latest_embedding: Optional[bool] = False,
+    ) -> None:
         if not isinstance(frame_id, str):
             raise Exception("frame ids must be strings")
 
@@ -75,6 +85,9 @@ class BaseLabelSet:
         self.crop_embeddings = []
         self._label_ids_set = set()
         self._seen_label_class_ids = set()
+        self.reuse_latest_embedding = (
+            False if reuse_latest_embedding is None else reuse_latest_embedding
+        )
 
     def add_crop_embedding(
         self, *, label_id: str, embedding: List[float], model_id: str = ""
@@ -114,7 +127,13 @@ class BaseLabelSet:
         """
         row = {
             "task_id": self.frame_id,
-            "label_data": self.label_data,
+            "label_data": [
+                {
+                    **cast(Dict[str, Any], label),
+                    "reuse_latest_embedding": self.reuse_latest_embedding,
+                }
+                for label in self.label_data
+            ],
             "type": self.crop_type,
         }
 
@@ -176,10 +195,14 @@ class BaseLabelSet:
         if not isinstance(classification, str):
             raise Exception("classifications must be strings")
 
+        for coord in [x1, y1, x2, y2]:
+            if not is_valid_number(coord):
+                raise Exception("all line segment coordinates must be valid int/float")
+
         attrs = {"x1": x1, "y1": y1, "x2": x2, "y2": y2}
 
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -249,6 +272,10 @@ class BaseLabelSet:
         if not isinstance(classification, str):
             raise Exception("classifications must be strings")
 
+        for dim in [top, left, width, height]:
+            if not is_valid_number(dim):
+                raise Exception("all bounding box dimensions must be valid int/float")
+
         attrs = {
             "top": top,
             "left": left,
@@ -257,7 +284,7 @@ class BaseLabelSet:
         }
 
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -268,6 +295,10 @@ class BaseLabelSet:
             attrs["iscrowd"] = iscrowd
         # TODO: This is mostly legacy for vergesense
         if area is not None:
+            if not is_valid_float(area):
+                raise Exception("area must be floats")
+            if area < 0.0:
+                raise Exception("area must be positive")
             attrs["area"] = area
 
         if links is not None:
@@ -329,7 +360,7 @@ class BaseLabelSet:
         }
 
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -402,8 +433,12 @@ class BaseLabelSet:
             "rot_w": rotation[3],
         }
 
+        for k, v in attrs.items():
+            if not is_valid_float(v):
+                raise Exception(k + " must be valid float")
+
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -475,6 +510,25 @@ class BaseLabelSet:
         if not isinstance(classification, str):
             raise Exception("classifications must be strings")
 
+        for dim in [top, left, width, height]:
+            if not is_valid_number(dim):
+                raise Exception("all bounding box dimensions must be valid int/float")
+
+        keypoint_names_set = set()
+        for keypoint in keypoints:
+            x = keypoint["x"]
+            y = keypoint["y"]
+            name = keypoint["name"]
+            if not is_valid_number(x):
+                raise Exception("keypoint x coordinate must be valid int/float")
+            if not is_valid_number(y):
+                raise Exception("keypoint y coordinate must be valid int/float")
+            if not isinstance(name, str) or len(name) == 0:
+                raise Exception("keypoint name must be non-zero length string")
+            if name in keypoint_names_set:
+                raise Exception("keypoint names must be unique within a label")
+            keypoint_names_set.add(name)
+
         attrs = {
             "top": top,
             "left": left,
@@ -484,7 +538,7 @@ class BaseLabelSet:
         }
 
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -553,10 +607,25 @@ class BaseLabelSet:
         if not isinstance(classification, str):
             raise Exception("classifications must be strings")
 
+        for polygon in polygons:
+            for vertex in polygon["vertices"]:
+                for num in vertex:
+                    if not is_valid_number(num):
+                        raise Exception(
+                            "all polygon vertex coordinates must be valid int/float"
+                        )
+
+        if center is not None:
+            for num in center:
+                if not is_valid_number(num):
+                    raise Exception(
+                        "polygon center coordinates must be valid int/float"
+                    )
+
         attrs: Dict[str, Any] = {"polygons": polygons, "center": center}
 
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -585,6 +654,7 @@ class BaseLabelSet:
         label_id: str,
         mask_url: Optional[str] = None,
         mask_data: Optional[Any] = None,
+        resize_mode: Optional[ResizeMode] = None,
     ) -> None:
         """
         Add an inference for 2D semseg. These should provide either a mask_url or a mask_data.
@@ -594,6 +664,9 @@ class BaseLabelSet:
             label_id (str): label_id which is unique across datasets and inferences.
             mask_url (str, optional): URL to the pixel mask png.
             mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
+            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
+                "fill": stretch the mask to fit the base image dimensions
+                None: no change
         """
 
         if not isinstance(label_id, str):
@@ -612,7 +685,7 @@ class BaseLabelSet:
                     "label": "__mask",
                     "label_coordinate_frame": sensor_id,
                     "label_type": "SEMANTIC_LABEL_URL_2D",
-                    "attributes": {"url": mask_url},
+                    "attributes": {"url": mask_url, "resize_mode": resize_mode},
                 }
             )
             self._label_ids_set.add(label_id)
@@ -661,7 +734,139 @@ class BaseLabelSet:
                     "label": "__mask",
                     "label_coordinate_frame": sensor_id,
                     "label_type": "SEMANTIC_LABEL_ASSET_2D",
-                    "attributes": {"asset_uuid": asset_uuid},
+                    "attributes": {
+                        "asset_uuid": asset_uuid,
+                        "resize_mode": resize_mode,
+                    },
+                }
+            )
+            self._label_ids_set.add(label_id)
+
+    # TODO: Do we want to expect column major or row major?
+    def add_2d_instance_seg(
+        self,
+        *,
+        sensor_id: str,
+        label_id: str,
+        mask_url: Optional[str] = None,
+        mask_data: Optional[Any] = None,
+        instance_mapping: List[Dict[str, Any]],
+        resize_mode: Optional[ResizeMode] = None,
+    ) -> None:
+        """
+        Add an label for 2D instance segmentation. These should provide either a mask_url or a mask_data, and a instance mapping.
+
+        Args:
+            sensor_id (str): sensor_id
+            label_id (str): label_id which is unique across datasets and inferences.
+            mask_url (str, optional): URL to the pixel mask png.
+            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
+            instance_mapping (List[Dict]): a list of instances present in the mask,
+                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
+                As an example of one instance:
+                .. code-block::
+                        {
+                            'id': 1,
+                            'classification': "Person",
+                            'attributes': {
+                                'is_standing': false,
+                            }
+                        }
+            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
+                "fill": stretch the mask to fit the base image dimensions
+                None: no change
+        """
+
+        if not isinstance(label_id, str):
+            raise Exception("label ids must be strings")
+        if "/" in label_id:
+            raise Exception("label ids cannot contain slashes (/)")
+
+        if mask_data is None and mask_url is None:
+            raise Exception("Please provide either a mask_data or a mask_url")
+
+        for instance_info in instance_mapping:
+            attrs = instance_info.get("attributes")
+            if attrs is None:
+                attrs = {}
+            attrs["id"] = instance_info.get("id")
+            attrs["resize_mode"] = resize_mode
+            self.label_data.append(
+                {
+                    "uuid": label_id + "_" + str(instance_info.get("id")),
+                    "linked_labels": [],
+                    "label": instance_info.get("classification", ""),
+                    "label_coordinate_frame": sensor_id,
+                    "label_type": "INSTANCE_LABEL_2D",
+                    "attributes": attrs,
+                }
+            )
+
+        if mask_url is not None:
+            self.label_data.append(
+                {
+                    "uuid": label_id,
+                    "linked_labels": [],
+                    "label": "__mask",
+                    "label_coordinate_frame": sensor_id,
+                    "label_type": "INSTANCE_LABEL_URL_2D",
+                    "attributes": {"url": mask_url, "resize_mode": resize_mode},
+                }
+            )
+            self._label_ids_set.add(label_id)
+
+        elif mask_data is not None:
+            if not HAS_PILLOW:
+                raise Exception(
+                    "To use mask_data, please install with additional image dependencies using `pip install aquariumlearning[img]`"
+                )
+
+            if not isinstance(mask_data, (np.ndarray, np.generic)):
+                raise Exception("mask_data must be a numpy ndarray")
+
+            if not np.issubdtype(mask_data.dtype, np.integer):
+                raise Exception("mask_data must have an integer dtype")
+
+            if len(mask_data.shape) != 2:
+                raise Exception(
+                    "mask_data must be a 2d ndarray of shape [height, width]"
+                )
+
+            # TODO: Validate
+            # TODO: Is png encoding too slow? Would something like pycocotools RLE be better?
+            unique_label_ids = np.unique(mask_data)
+            for v in unique_label_ids:
+                self._seen_label_class_ids.add(cast(int, v))
+
+            # Compress to a png for later usage, and immediate compression
+            png_img = Image.fromarray(mask_data.astype("uint8"))
+            png_io = BytesIO()
+            png_img.save(
+                png_io, format="PNG", compress_level=6
+            )  # 0-9, defualt is 6, we favor slightly faster
+            png_io.seek(0)
+            png_bytes = png_io.read()
+
+            asset_uuid = str(uuid4())
+            self.label_assets.append(
+                {
+                    "uuid": asset_uuid,
+                    "asset_type": "INSTANCE_SEG_MASK",
+                    "value": png_bytes,
+                }
+            )
+
+            self.label_data.append(
+                {
+                    "uuid": label_id,
+                    "linked_labels": [],
+                    "label": "__mask",
+                    "label_coordinate_frame": sensor_id,
+                    "label_type": "INSTANCE_LABEL_ASSET_2D",
+                    "attributes": {
+                        "asset_uuid": asset_uuid,
+                        "resize_mode": resize_mode,
+                    },
                 }
             )
             self._label_ids_set.add(label_id)
@@ -697,7 +902,7 @@ class BaseLabelSet:
         attrs = {}
 
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -755,7 +960,7 @@ class BaseLabelSet:
         attrs = {}
 
         if confidence is not None:
-            if not isinstance(confidence, float):
+            if not is_valid_float(confidence):
                 raise Exception("confidence must be floats")
             if confidence < 0.0 or confidence > 1.0:
                 raise Exception("confidence must be between 0.0 and 1.0")
@@ -780,20 +985,38 @@ class BaseLabelSet:
 class InferenceLabelSet(BaseLabelSet):
     custom_metrics: Dict[str, CustomMetricsEntry]
     visualization_assets: Dict[str, str]
+    user_metadata: Dict[str, Any]
+    reuse_latest_embedding: bool
 
-    def __init__(self, *, frame_id: str) -> None:
-        super().__init__(frame_id=frame_id, crop_type="Inference")
+    def __init__(
+        self, *, frame_id: str, reuse_latest_embedding: Optional[bool] = False
+    ) -> None:
+        super().__init__(
+            frame_id=frame_id,
+            crop_type="Inference",
+            reuse_latest_embedding=reuse_latest_embedding,
+        )
         self.custom_metrics = {}
         self.visualization_assets = {}
+        self.user_metadata = {}
+        self.reuse_latest_embedding = (
+            False if reuse_latest_embedding is None else reuse_latest_embedding
+        )
 
     def add_custom_metric(self, name: str, value: CustomMetricsEntry) -> None:
         """Add a custom metric for a given inference frame.
 
         Args:
             name (str): The name of the custom metric being added. Must match one of the custom_metrics already defined by the corresponding Project.
-            value (Union[float, List[Union[int, float]]]): The value of the custom metric (either a float or 2d list of floats/integers).
+            value (Union[float, List[List[Union[int, float]]]]): The value of the custom metric (either a float or 2d list of floats/integers).
         """
-        if not (isinstance(value, float) or isinstance(value, list)):
+        if not (
+            is_valid_float(value)
+            or (
+                isinstance(value, list)
+                and all(all(is_valid_number(y) for y in x) for x in value)
+            )
+        ):
             raise Exception(
                 "Custom metrics values must be either a float, or a 2D list of floats/integers."
             )
@@ -803,14 +1026,38 @@ class InferenceLabelSet(BaseLabelSet):
     def add_visualization_asset(self, image_id: str, image_url: str) -> None:
         self.visualization_assets[image_id] = image_url
 
+    def add_user_metadata(
+        self,
+        key: str,
+        val: Union[str, int, float, bool],
+    ) -> None:
+        """Add a user provided metadata field.
+
+        The types of these metadata fields will be infered and they'll be made
+        available in the app for querying and metrics.
+
+        Args:
+            key (str): The key for your metadata field
+            val (Union[str, int, float, bool]): value
+            val_type (Literal["str", "int", "float", "bool"], optional): type of val as string. Defaults to None.
+        """
+        assert_valid_name(key)
+        # Validates that neither val or type is None
+        self.user_metadata[key] = val
+
     def to_dict(self) -> Dict[str, Any]:
         row = super().to_dict()
 
         row["custom_metrics"] = self.custom_metrics
-        row["inference_metadata"] = {"visualization_assets": self.visualization_assets}
+        row["inference_metadata"] = {
+            "visualization_assets": self.visualization_assets,
+        }
         row["inference_data"] = row["label_data"]
+        row["reuse_latest_embedding"] = self.reuse_latest_embedding
         del row["label_data"]
 
+        row["inference_metadata"]["user_metadata"] = self.user_metadata
+
         return row
 
     def _to_summary(self) -> InferenceFrameSummary:
@@ -825,8 +1072,14 @@ class InferenceLabelSet(BaseLabelSet):
 
 
 class GTLabelSet(BaseLabelSet):
-    def __init__(self, *, frame_id: str) -> None:
-        super().__init__(frame_id=frame_id, crop_type="GT")
+    def __init__(
+        self, *, frame_id: str, reuse_latest_embedding: Optional[bool] = False
+    ) -> None:
+        super().__init__(
+            frame_id=frame_id,
+            crop_type="GT",
+            reuse_latest_embedding=reuse_latest_embedding,
+        )
 
 
 class UpdateGTLabelSet(GTLabelSet):
